// AutoCloseable.hpp

#ifndef BRYNHILD_CORE_AUTOCLOSEABLE_HPP
#define BRYNHILD_CORE_AUTOCLOSEABLE_HPP

namespace brynhild {
    namespace core {

        /**
         * An object that may hold resources (such as file or socket handles)
         * until it is closed. The close method of an AutoCloseable
         * object should be called to release the resources.
         *
         * This is similar to Java's AutoCloseable interface.
         */
        class AutoCloseable {
        public:
            /**
             * Virtual destructor. Subclasses should override this method
             * to release any underlying resources.
             */
            virtual ~AutoCloseable() {}

            /**
             * Closes this resource, relinquishing any underlying resources.
             * This method should be overridden by subclasses to perform
             * resource release.
             */
            virtual void close() = 0;
        };

    } // namespace core
} // namespace brynhild

#endif // BRYNHILD_CORE_AUTOCLOSEABLE_HPP
