// FileInputStream.hpp

#ifndef BRYNHILD_IO_FILEINPUTSTREAM_HPP
#define BRYNHILD_IO_FILEINPUTSTREAM_HPP

#include <fstream>
#include <string>
#include <memory>
#include <mutex>
#include "InputStream.hpp"
#include "IOException.hpp"

namespace brynhild {
    namespace io {

        class FileInputStream : public InputStream {
        private:
            std::unique_ptr<std::ifstream> fileStream;
            std::string filePath;
            std::mutex mutex;
            std::streampos markPosition = 0;  // Position marker
            bool isMarkSet = false;           // Flag to check if mark is set

        public:
            // Constructor with file path
            explicit FileInputStream(const std::string& path)
                : filePath(path), fileStream(std::make_unique<std::ifstream>(path, std::ios::binary)) {
                if (!fileStream->is_open()) {
                    throw IOException("File not found: " + path);
                }
            }

            // Override read method from InputStream
            int read() override {
                std::lock_guard<std::mutex> lock(mutex); // Lock for thread safety
                char ch;
                fileStream->get(ch);
                if (fileStream->eof()) {
                    return -1;  // End of file
                }
                if (!fileStream->good()) {
                    throw IOException("Read error");
                }
                return static_cast<unsigned char>(ch);  // Return byte as an unsigned char converted to int
            }

            // Override close method from InputStream
            void close() override {
                std::lock_guard<std::mutex> lock(mutex); // Lock for thread safety
                if (fileStream && fileStream->is_open()) {
                    fileStream->close();
                }
            }

            // Mark the current position in the input stream
            void mark(int readLimit) override {
                std::lock_guard<std::mutex> lock(mutex); // Lock for thread safety
                markPosition = fileStream->tellg();
                isMarkSet = true;
            }

            // Reset the stream to the most recent mark
            void reset() override {
                std::lock_guard<std::mutex> lock(mutex); // Lock for thread safety
                if (!isMarkSet) {
                    throw IOException("Mark not set");
                }
                fileStream->seekg(markPosition);
            }

            // Check if mark and reset are supported
            bool markSupported() const override {
                return true;  // Mark and reset are supported
            }

            // Destructor
            ~FileInputStream() {
                close();
            }
        };

    } // namespace io
} // namespace brynhild

#endif // BRYNHILD_IO_FILEINPUTSTREAM_HPP
