// InputStream.hpp

#ifndef BRYNHILD_CORE_INPUTSTREAM_HPP
#define BRYNHILD_CORE_INPUTSTREAM_HPP

#include <iostream>
#include "Closeable.hpp"
#include "IOException.hpp"

namespace brynhild {
    namespace io {

        class InputStream : public Closeable {
        public:
            // Virtual destructor to ensure proper cleanup in derived classes
            virtual ~InputStream() {
                close();
            }

            // Pure virtual function to read the next byte of data
            virtual int read() = 0;

            // Reads a sequence of bytes into a buffer
            virtual int read(char* buffer, int offset, int length) {
                if (buffer == nullptr) {
                    throw IOException("Buffer is null");
                }
                if (offset < 0 || length < 0 || length > INT_MAX - offset) {
                    throw IOException("Invalid offset or length");
                }

                int bytesRead = 0;
                for (int i = 0; i < length; ++i) {
                    int byte = read();
                    if (byte == -1) {
                        return bytesRead > 0 ? bytesRead : -1;
                    }
                    buffer[offset + i] = static_cast<char>(byte);
                    bytesRead++;
                }
                return bytesRead;
            }

            // Skips over and discards n bytes of data
            virtual long skip(long n) {
                if (n <= 0) {
                    return 0;
                }

                int skippedBytes = 0;
                for (long i = 0; i < n; ++i) {
                    if (read() == -1) {
                        break;
                    }
                    skippedBytes++;
                }
                return skippedBytes;
            }

            // Returns an estimate of the number of bytes that can be read
            virtual int available() const {
                return 0;  // Default implementation, subclasses may override
            }

            // Closes this input stream
            virtual void close() override {
                // Implementation depends on the specific subclass
            }

            // Marks the current position in this input stream
            virtual void mark(int readLimit) {
                // Default implementation does nothing
            }

            // Repositions this stream to the position at the time the mark method was last called
            virtual void reset() {
                throw IOException("mark/reset not supported");
            }

            // Tests if this input stream supports the mark and reset methods
            virtual bool markSupported() const {
                return false;  // Default implementation
            }
        };

    } // namespace core
} // namespace brynhild

#endif // BRYNHILD_CORE_INPUTSTREAM_HPP
