// Reader.hpp

#ifndef BRYNHILD_IO_READER_HPP
#define BRYNHILD_IO_READER_HPP

#include <mutex>
#include <stdexcept>
#include "IOException.hpp"
#include "../core/Readable.hpp"
#include "Closeable.hpp"

namespace brynhild {
    namespace io {

        class Reader : public core::Readable, public Closeable {
        protected:
            std::mutex lock; // Mutex used for synchronization within the class

        public:
            Reader() = default; // Default constructor
            virtual ~Reader() = default; // Default destructor

            // Pure virtual methods to be implemented by subclasses
            virtual int read(char* cbuf, int off, int len) = 0;
            virtual void close() = 0;

            // Other virtual methods with default implementations
            virtual int read(char* cbuf, int len) {
                std::lock_guard<std::mutex> guard(lock); // Locking for thread safety
                return read(cbuf, 0, len);
            }
            virtual int read() {
                std::lock_guard<std::mutex> guard(lock); // Locking for thread safety
                char cb[1];
                int result = read(cb, 0, 1);
                return (result == -1) ? -1 : static_cast<int>(cb[0]);
            }
            long skip(long n){
                if (n <= 0) {
                    return 0;
                }

                long skipped = 0;
                for (long i = 0; i < n; ++i) {
                    if (read() == -1) {
                        break;
                    }
                    skipped++;
                }
                return skipped;
            }
            virtual bool ready() {
                std::lock_guard<std::mutex> guard(lock); // Locking for thread safety
                return false;
            }
            virtual bool markSupported() {
                std::lock_guard<std::mutex> guard(lock); // Locking for thread safety
                return false;
            }
            virtual void mark(int readAheadLimit) {
                std::lock_guard<std::mutex> guard(lock); // Locking for thread safety
                throw IOException("mark() not supported");
            }
            virtual void reset() {
                std::lock_guard<std::mutex> guard(lock); // Locking for thread safety
                throw IOException("reset() not supported");
            }
        };

    } // namespace io
} // namespace brynhild

#endif // BRYNHILD_IO_READER_HPP
